/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */

#include <sys/epoll.h>
#include <sys/socket.h>
#ifndef EPOLLRDHUP
#  include <linux/poll.h>
#  ifdef POLLRDHUP
#    define EPOLLRDHUP POLLRDHUP
#  else
#    define EPOLLRDHUP 0
#  endif
#endif
#include "evtloop.h"
#include "mutt.h"

static int epollfd = -1;

static int el_job_setemode(job_t *w, el_mode emode)
{
    static int const evtmode_to_epoll[] = {
        [EL_NEW]     = EPOLLRDHUP,
        [EL_READING] = EPOLLIN,
        [EL_WRITING] = EPOLLOUT,
        [EL_RDWR]    = EPOLLIN | EPOLLOUT,
        [EL_IDLE]    = EPOLLRDHUP,
    };

    assert (w->mode == emode || emode == EL_WRITING || emode == EL_READING);

    if (emode != w->emode) {
        struct epoll_event event = {
            .data.ptr = w,
            .events   = evtmode_to_epoll[emode],
        };
        int action = w->emode == EL_NEW ? EPOLL_CTL_ADD : EPOLL_CTL_MOD;
        if (epoll_ctl(epollfd, action, w->fd, &event) < 0) {
            return el_job_release(w, true);
        }
    }
    w->emode = emode;
    return 0;
}

int el_job_setmode(job_t *w, el_mode mode)
{
    if (w->mode == w->emode) {
        w->mode = mode;
        return el_job_setemode(w, mode);
    } else {
        w->mode = mode;
        return 0;
    }
}

int el_job_release(job_t *w, el_status reason)
{
    w->state = EL_LLP_FINI;
    if (w->m && w->m->finalize) {
        w->m->finalize(w, reason);
    }
    if (w->fd >= 0) {
        close(w->fd);
    }
    job_delete(&w);
    return -1;
}

static int el_job_connecting(job_t *w)
{
    int err = 0;
    socklen_t len = sizeof(err);

    if (getsockopt(w->fd, SOL_SOCKET, SO_ERROR, (void *)&err, &len) || err)
        return el_job_release(w, EL_ERROR);

    w->state = EL_LLP_READY;
    return w->m->on_event(w, EL_EVT_RUNNING);
}

int el_job_connect(job_t *w, struct sockaddr *addr, socklen_t len,
                   int type, int proto)
{
    int res, sock = socket(addr->sa_family, type, proto);

    if (sock < 0)
        goto error;

    res = fcntl(sock, F_GETFL);
    if (res < 0)
        goto error;
    if (fcntl(sock, F_SETFL, res | O_NONBLOCK) < 0)
        goto error;
    if (connect(sock, addr, len) < 0)
        goto error;

    w->fd  = sock;
    w->llp = &el_job_connecting;
    return el_job_setmode(w, EL_WRITING);

  error:
    close(sock);
    return el_job_release(w, EL_ERROR);
}

int el_dispatch(int timeout)
{
    struct epoll_event events[FD_SETSIZE];
    int count = epoll_wait(epollfd, events, countof(events), timeout);

    if (count < 0) {
        if (errno == EAGAIN || errno == EINTR)
            return 0;
        mutt_error("epoll_wait");
        mutt_exit(EXIT_FAILURE);
    }

    while (--count >= 0) {
        job_t *w  = events[count].data.ptr;
        int event = events[count].events;
        int evt   = 0;

        switch (w->state) {
          case EL_LLP_INIT:
            w->llp(w);
            break;

          case EL_LLP_READY:
            if (event & EPOLLRDHUP) {
                IGNORE(el_job_release(w, EL_RDHUP));
            } else if (w->mode != w->emode) {
                    w->m->on_event(w, EL_EVT_INOUT ^ w->emode);
            } else {
                if (event & EPOLLIN)
                    evt |= EL_EVT_IN;
                if (event & EPOLLOUT)
                    evt |= EL_EVT_OUT;
                w->m->on_event(w, evt);
            }
            break;

          default:
            IGNORE(el_job_release(w, EL_ERROR));
            break;
        }
    }

    return 0;
}
